@@ -24,8 +24,8 @@ import json5
 class TestLoads(unittest.TestCase):
     maxDiff = None
 
-    def check(self, s, obj):
-        self.assertEqual(json5.loads(s), obj)
+    def check(self, s, obj, strict=True):
+        self.assertEqual(obj, json5.loads(s, strict=strict))
 
     def check_fail(self, s, err=None):
         try:
@@ -51,9 +51,6 @@ class TestLoads(unittest.TestCase):
         self.check('true', True)
         self.check('false', False)
 
-    def test_cls_is_not_supported(self):
-        self.assertRaises(AssertionError, json5.loads, '1', cls=lambda x: x)
-
     def test_duplicate_keys_should_be_allowed(self):
         self.assertEqual(
             json5.loads('{foo: 1, foo: 2}', allow_duplicate_keys=True),
@@ -85,12 +82,17 @@ class TestLoads(unittest.TestCase):
 
     def test_encoding(self):
         self.assertEqual(json5.loads(b'"\xf6"', encoding='iso-8859-1'), '\xf6')
+        with self.assertRaises(UnicodeDecodeError):
+            json5.load(io.BytesIO(b'\xff'))
 
     def test_numbers(self):
         # decimal literals
         self.check('1', 1)
         self.check('-1', -1)
         self.check('+1', 1)
+        self.check_fail('--6', '<string>:1 Unexpected "-" at column 2')
+        self.check_fail('++6', '<string>:1 Unexpected "+" at column 2')
+        self.check_fail('+-6', '<string>:1 Unexpected "-" at column 2')
 
         # hex literals
         self.check('0xf', 15)
@@ -104,6 +106,9 @@ class TestLoads(unittest.TestCase):
         self.check('1.5', 1.5)
         self.check('1.5e3', 1500.0)
         self.check('-0.5e-2', -0.005)
+        self.check_fail('--6.0', '<string>:1 Unexpected "-" at column 2')
+        self.check_fail('++6.0', '<string>:1 Unexpected "+" at column 2')
+        self.check_fail('+-6.0', '<string>:1 Unexpected "-" at column 2')
 
         # names
         self.check('Infinity', float('inf'))
@@ -234,6 +239,22 @@ class TestLoads(unittest.TestCase):
             obj,
         )
 
+    def test_load_fails(self):
+        s = io.StringIO('x')
+        with self.assertRaises(ValueError):
+            json5.load(s)
+
+    def test_strict(self):
+        # From [GitHub issue #82](https://github.com/dpranke/pyjson5/issues/82)
+        d = '{\n"key": "value\nover two lines",\n}'
+        self.check_fail(d, '<string>:2 Unexpected "\n" at column 14')
+        self.check(d, {'key': 'value\nover two lines'}, strict=False)
+
+        # Test w/ single quotes to get coverage.
+        d = "{\n'key': 'value\nover two lines',\n}"
+        self.check_fail(d, '<string>:2 Unexpected "\n" at column 14')
+        self.check(d, {'key': 'value\nover two lines'}, strict=False)
+
     def test_strings(self):
         self.check('"foo"', 'foo')
         self.check("'foo'", 'foo')
@@ -313,6 +334,47 @@ class TestLoads(unittest.TestCase):
         self.check_fail('0 a', '<string>:1 Unexpected "a" at column 3')
 
 
+class TestParse(unittest.TestCase):
+    maxDiff = None
+
+    def check(self, s, obj, err, pos, consume_trailing=True, start=0):
+        actual_obj, actual_err, actual_pos = json5.parse(
+            s, consume_trailing=consume_trailing, start=start
+        )
+        self.assertEqual(obj, actual_obj)
+        self.assertEqual(err, actual_err)
+        self.assertEqual(pos, actual_pos)
+
+    def test_parse(self):
+        self.check('4', 4, err=None, pos=1)
+        self.check('4 ', 4, err=None, pos=2)
+        self.check('4 ', 4, err=None, pos=1, consume_trailing=False)
+        self.check(
+            'x', None, err='<string>:1 Unexpected "x" at column 1', pos=0
+        )
+        self.check(
+            '4 x', None, err='<string>:1 Unexpected "x" at column 3', pos=2
+        )
+        self.check('4 x', 4, err=None, pos=1, consume_trailing=False)
+        self.check('x 4', 4, err=None, pos=3, start=1)
+        self.check(
+            'x y',
+            None,
+            err='<string>:1 Unexpected "y" at column 3',
+            pos=2,
+            start=1,
+        )
+
+        val, err, pos = json5.parse('[][]', consume_trailing=False)
+        self.assertEqual(val, [])
+        self.assertEqual(err, None)
+        self.assertEqual(pos, 2)
+        val, err, pos = json5.parse('[][]', consume_trailing=False, start=pos)
+        self.assertEqual(val, [])
+        self.assertEqual(err, None)
+        self.assertEqual(pos, 4)
+
+
 class TestDump(unittest.TestCase):
     def test_basic(self):
         sio = io.StringIO()
@@ -323,8 +385,8 @@ class TestDump(unittest.TestCase):
 class TestDumps(unittest.TestCase):
     maxDiff = None
 
-    def check(self, obj, s):
-        self.assertEqual(json5.dumps(obj), s)
+    def check(self, obj, s, **kwargs):
+        self.assertEqual(s, json5.dumps(obj, **kwargs))
 
     def test_allow_duplicate_keys(self):
         self.assertIn(
@@ -351,18 +413,25 @@ class TestDumps(unittest.TestCase):
         self.check(False, 'false')
 
     def test_check_circular(self):
+        # This tests that a non-cyclic object works w/ either flag value.
+        obj = [1, 2, 3]
+        self.check(obj, '[1, 2, 3]')  # testing the default
+        self.check(obj, '[1, 2, 3]', check_circular=True)
+        self.check(obj, '[1, 2, 3]', check_circular=False)
+
         # This tests a trivial cycle.
         obj = [1, 2, 3]
         obj[2] = obj
         self.assertRaises(ValueError, json5.dumps, obj)
 
-        # This checks that json5 doesn't raise an error. However,
+        # This checks that json5 doesn't raise an error with
+        # check_circular=false and a cycle. However,
         # the underlying Python implementation likely will.
         try:
             json5.dumps(obj, check_circular=False)
             self.fail()  # pragma: no cover
-        except Exception as e:
-            self.assertNotIn(str(e), 'Circular reference detected')
+        except RecursionError:
+            pass
 
         # This checks that repeated but non-circular references
         # are okay.
@@ -511,6 +580,13 @@ class TestDumps(unittest.TestCase):
     def test_null(self):
         self.check(None, 'null')
 
+    def test_separators(self):
+        # Check that custom separators work; these separators add an
+        # extra space.
+        self.check(
+            [{'foo': 1}, 2], '[{foo:  1},  2]', separators=(',  ', ':  ')
+        )
+
     def test_objects(self):
         self.check({'foo': 1}, '{foo: 1}')
         self.check({'foo bar': 1}, '{"foo bar": 1}')
@@ -538,19 +614,189 @@ class TestDumps(unittest.TestCase):
             json5.dumps({'foo': 1}, quote_keys=True), '{"foo": 1}'
         )
 
-    def test_strings(self):
+    def test_strings_containing_backslashes_and_quotes(self):
+        # Understanding how things are escaped in test of JSON5 strings
+        # can be tricky.
+
+        # Normal Python escaping means that the following asserts are true:
+        self.assertEqual('\\z', '\\z')
+        self.assertEqual('\\z', r'\z')
+        self.assertEqual('\\z', r'\z')
+
+        # But, in Python, escaping quotes in a raw string is tricky, because
+        # the escape is left in the output. The results of this are:
+
+        # (1) You cannot use a raw string to match a value ending in a
+        # an odd number of backslashes: the first N-1 backslashes would
+        # be matched by the same number of backslashes in the raw string,
+        # leaving a single backslash followed by a quote. The quote
+        # would then be considered escaped, leaving the string unterminated.
+        # Ending in an even number of backslashes is fine:
+        self.assertEqual(len(r'\\'), 2)
+        self.assertEqual(r'\\', '\\\\')
+
+        # (2) You cannot use a raw string to represent a value that contains
+        # the same kind of quote you're using to enclose the string, unless the
+        # value actually contains an odd number of backslashes immediately
+        # preceding the quote:
+        self.assertEqual(len(r'\''), 2)
+        self.assertEqual(r'\'', "\\'")
+        self.assertEqual(r'\'', "\\'")
+        self.assertEqual(r'\'', r'\'')
+
+        # Now, in JSON5, if the value doesn't contain backslashes, you can
+        # use normal quoting as you would in Python, but you can't use
+        # raw strings, since the raw strings would require the values to
+        # have backslashes in them:
         self.check("'single'", '"\'single\'"')
-        self.check('"double"', '"\\"double\\""')
-        self.check(
-            "'single \\' and double \"'", '"\'single \\\\\' and double \\"\'"'
-        )
+        self.check("'single'", '"\'single\'"')
+
+        # In order to represent a backslash in the value you also need to
+        # escape it in the JSON string: a string containing a single backslash
+        # is represented by "\\". So, in order to match that single backslash
+        # via non-raw strings in Python source code, you need to (3) double the
+        # backslashes (for JSON5) and then double them again, for python source
+        # code. I.e., you need *4* backslashes in the source code. In many
+        # cases you can also use single-quoted raw strings (where you have
+        # to (4) double the number of quotes in the output), but in this
+        # particular example, you cannot use single-quoted raw strings,
+        # due to (1).
+        self.check('\\', '"\\\\"')
+        self.check('\\', '"\\\\"')
+
+        # You cannot use a double-quoted raw string to represent
+        # double-quoted JSON5 strings, since the output needs to start with a
+        # double quote, and you can't represent that in a raw double-quoted
+        # string due to (2).
+
+        # Here you see the doubling of backslashes in single-quoted
+        # raw output strings, and the quadrupling in a non-raw string.
+        self.check(r'\z', r'"\\z"')
+        self.check(r'\\z', r'"\\\\z"')
+        self.check(r'\\\z', r'"\\\\\\z"')
+        self.check(r'\z', '"\\\\z"')
+
+        self.check('"', '"\\""')
+        self.check('"', '"\\""')
+
+        # Here it's okay to use a raw string for output since the output
+        # needs to have a single backslash and doesn't end in a single quote.
+        self.check('"', r'"\""')
+
+        # Here you cannot use raw strings for the output as the output
+        # would need to have only two backslashes in it.
+        self.check(r'\'', '"\\\\\'"')
+        self.check(r'\'', '"\\\\\'"')
 
     def test_string_escape_sequences(self):
+        # self.check(r'\'', '"\\\\\'"')
+        self.check("'\\'", '"\'\\\\\'"')
         self.check(
             '\u2028\u2029\b\t\f\n\r\v\\\0',
-            '"\\u2028\\u2029\\b\\t\\f\\n\\r\\v\\\\\\0"',
+            r'"\u2028\u2029\b\t\f\n\r\v\\\0"',
         )
 
+    def test_string_quote_styles(self):
+        def checkp(**kwargs):
+            return lambda obj, s: self.assertEqual(
+                s, json5.dumps(obj, **kwargs)
+            )
+
+        sq = "'"
+        dq = '"'
+
+        neither = 'a b c'
+        single_neither = sq + neither + sq
+        double_neither = dq + neither + dq
+
+        single = "a 'b' c"
+        single_single = sq + r'a \'b\' c' + sq
+        double_single = dq + single + dq
+
+        double = 'a "b" c'
+        single_double = sq + double + sq
+        double_double = dq + r'a \"b\" c' + dq
+
+        both = 'a \'b\' "c" d'
+        single_both = sq + r'a \'b\' "c" d' + sq
+        double_both = dq + r"a 'b' \"c\" d" + dq
+
+        reverse = 'a "b" \'c\' d'
+        single_reverse = sq + r'a "b" \'c\' d' + sq
+        double_reverse = dq + r"a \"b\" 'c' d" + dq
+
+        more_single = "a 'b' 'c' \"d\" e"
+        single_more_single = sq + r'a \'b\' \'c\' "d" e' + sq
+        double_more_single = dq + r"a 'b' 'c' \"d\" e" + dq
+
+        more_double = 'a "b" "c" \'d\' e'
+        single_more_double = sq + r'a "b" "c" \'d\' e' + sq
+        double_more_double = dq + r"a \"b\" \"c\" 'd' e" + dq
+
+        more_single_double_first = "a \"b\" 'c' 'd' e"
+        single_more_single_double_first = sq + r'a "b" \'c\' \'d\' e' + sq
+        double_more_single_double_first = dq + r"a \"b\" 'c' 'd' e" + dq
+
+        more_double_single_first = 'a \'b\' "c" "d" e'
+        single_more_double_single_first = sq + r'a \'b\' "c" "d" e' + sq
+        double_more_double_single_first = dq + r"a 'b' \"c\" \"d\" e" + dq
+
+        # Default settings (should be ALWAYS_DOUBLE)
+        c = checkp()
+        c(neither, double_neither)
+        c(single, double_single)
+        c(double, double_double)
+        c(both, double_both)
+        c(reverse, double_reverse)
+        c(more_single, double_more_single)
+        c(more_double, double_more_double)
+        c(more_single_double_first, double_more_single_double_first)
+        c(more_double_single_first, double_more_double_single_first)
+
+        c = checkp(quote_style=json5.QuoteStyle.ALWAYS_DOUBLE)
+        c(neither, double_neither)
+        c(single, double_single)
+        c(double, double_double)
+        c(both, double_both)
+        c(reverse, double_reverse)
+        c(more_single, double_more_single)
+        c(more_double, double_more_double)
+        c(more_single_double_first, double_more_single_double_first)
+        c(more_double_single_first, double_more_double_single_first)
+
+        c = checkp(quote_style=json5.QuoteStyle.ALWAYS_SINGLE)
+        c(neither, single_neither)
+        c(single, single_single)
+        c(double, single_double)
+        c(both, single_both)
+        c(reverse, single_reverse)
+        c(more_single, single_more_single)
+        c(more_double, single_more_double)
+        c(more_single_double_first, single_more_single_double_first)
+        c(more_double_single_first, single_more_double_single_first)
+
+        c = checkp(quote_style=json5.QuoteStyle.PREFER_DOUBLE)
+        c(neither, double_neither)
+        c(single, double_single)
+        c(double, single_double)
+        c(both, double_both)
+        c(reverse, double_reverse)
+        c(more_single, double_more_single)
+        c(more_double, single_more_double)
+        c(more_single_double_first, double_more_single_double_first)
+        c(more_double_single_first, single_more_double_single_first)
+
+        c = checkp(quote_style=json5.QuoteStyle.PREFER_SINGLE)
+        c(neither, single_neither)
+        c(single, double_single)
+        c(double, single_double)
+        c(both, single_both)
+        c(reverse, single_reverse)
+        c(more_single, double_more_single)
+        c(more_double, single_more_double)
+        c(more_single_double_first, double_more_single_double_first)
+        c(more_double_single_first, single_more_double_single_first)
+
     def test_skip_keys(self):
         od = OrderedDict()
         od[(1, 2)] = 2
